<'
# Copyright (c) 2011-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Tue Dec 13 2011
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#ifndef H_GROS_<"$COMP">_INTERNALS
#define H_GROS_<"$COMP">_INTERNALS

#include <pthread.h>

#include "boost/shared_ptr.hpp"

#include "ros/ros.h"
#include "actionlib/server/action_server.h"

#include "<"$comp">_port.h"
#include "<"$comp">_remote.h"
#include "<"$comp">_service.h"
#include "<"$comp">_action.h"
#include "<"$comp">_locals.h"
#include "<"$comp">_activities.h"

/* service id */
<'set i 0; foreach s [$component services] {'>
#define <"$COMP">_<"[$s name]">_RQSTID	(<"$i">)
<' incr i '>
<'}'>
#define <"$COMP">_NRQSTID	(<"$i">)


/* --- context ------------------------------------------------------------- */

struct genom_component_data;

struct genom_context_data {
  genom_component_data *self;

  genom_event ex;
  void *exdetail;
  size_t exsize;
};


/* --- tasks --------------------------------------------------------------- */

struct genom_exec_task {
  genom_context_iface context;
  genom_context_data context_data;

  pthread_t id;
  int stateid;
  pthread_cond_t signal;
  bool runnable, wakeup;
  genom_event status, pact;

  void	spawn(genom_component_data *self,
              const char *name, void *(*start)(void *));
};

int	<"$comp">_cntrl_task(const char *instance);
<'foreach t [$component tasks] {'>
void *	<"$comp">_<"[$t name]">_task(void *data);
<'}'>


/* --- internal state ------------------------------------------------------ */

struct genom_component_data {
  genom_context_iface context;
  genom_context_data context_data;

  ros::NodeHandle *node;
  ros::Timer clock;
  unsigned int ticks;
  pthread_mutex_t task_lock;

<'if {![catch {$component ids} ids]} {'>
  <"[$ids declarator ids]">;
<'}'>

  struct ports {
<'foreach p [$component ports] {'>
    genom::<"$comp">_port::<"[$p name]"> <"[$p name]">;
<'}'>

<'set i [list]; foreach p [$component ports] { lappend i "[$p name](self)" } '>
   ports(genom_component_data *self) <"[expr {[llength $i]?":":""}]">
    <"[join $i ", "]">
      {}
  } ports;

  int genom_state_init(void);

  /* tasks */
  struct {
<'foreach t [$component tasks] {'>
    genom_exec_task <"[$t name]">;
<'}'>
  } tasks;

  void exec_task_signal(void);

  struct {
<'foreach r [$component remotes] {'>
    genom::<"$comp">_remote::<"[$r name]"> <"[$r name]">;
<'}'>
  } remotes;

/* service callbacks */
  struct {
<'foreach s [$component services] {'>
<'  if {[catch {$s task} t]} {'>
    ros::ServiceServer <"[$s name]">;
<'  } else {'>
    actionlib::ActionServer< genom::action::<"[$s name]"> > *<"[$s name]">;
<'  }'>
<'}'>
  } services;

<'foreach s [$component services] {'>
  genom_event <"[$s name]">_controlcb(
    genom::srv::<"[$s name]">::locals &locals,
    genom::srv::<"[$s name]">::input &in,
    genom::srv::<"[$s name]">::output &out);
<'  if {[catch {$s task} t]} {'>
  bool <"[$s name]">_rqstcb(
    genom::srv::<"[$s name]">::input &in,
    genom::srv::<"[$s name]">::output &out);
<'  } else {'>
  void <"[$s name]">_rqstcb(
    actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst);
  void <"[$s name]">_intercb(
    actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst);
<'  }'>

<'}'>

  /* activities */
  genom_activities activities;
  bool run_map[<"$COMP">_NRQSTID];

  void	cntrl_task_activitycb(genom_activity *a);

<'if {![catch {$component clock-rate} rate]} {'>
  /* ROS timer callback */
  void timercb(const ros::TimerEvent &e);
<'}'>

  genom_component_data() : ports(this) {}
};

/* global shutdown flag */
extern int genom_shutdown;

/* exceptions */
extern "C" {
genom_event	ros_server_raise(genom_event ex, void *detail,
                        size_t size, genom_context self);
const void *	ros_server_raised(genom_event *ex,
                        genom_context self);
}

/* log functions */
void	genom_log_info(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));
void	genom_log_warn(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));
void	genom_log_debug(const char *format, ...)
  __attribute__ ((format (printf, 1, 2)));

#endif /* H_GROS_<"$COMP">_INTERNALS */
